"""
辅助函数模块
"""
import time
import numpy as np
import pandas as pd


def run_time(func):
    def call_fun(*args, **kwargs):
        start_time = time.time()
        func(*args, **kwargs)
        end_time = time.time()
        print('程序用时：%s秒' % (end_time - start_time))
    return call_fun


def cal_annual_return(return_series):
    T = len(return_series)
    annual_return = (1 + return_series).prod() ** (52 / T) - 1

    return annual_return


def cal_annual_volatility(return_series):
    vol = return_series.std() * np.sqrt(52)

    return vol


def cal_max_drawdown(nav_series):
    drawdown_series = nav_series / (nav_series.cummax()) - 1

    return drawdown_series.min()

def cal_dd_recovery(nav_series):
    # drawdown = nav_series / (nav_series.cummax()) - 1

    time_to_recovery = pd.Series(index=nav_series.index)
    previous_peaks = pd.Series(index=nav_series.index)
    for i in range(len(nav_series)):
        if i == 0:
            previous_peaks.iloc[i] = nav_series.iloc[i]
            time_to_recovery.iloc[i] = 0
        else:
            if nav_series.iloc[i] > previous_peaks.iloc[i - 1]:
                previous_peaks.iloc[i] = nav_series.iloc[i]
                time_to_recovery.iloc[i] = 0
            else:
                previous_peaks.iloc[i] = previous_peaks.iloc[i - 1]
                time_to_recovery.iloc[i] = time_to_recovery.iloc[i - 1] + 1

    # max_drawdown_recovery_period = time_to_recovery[drawdown == drawdown.min()].max()
    max_drawdown_recovery_period = time_to_recovery.max()

    return max_drawdown_recovery_period


def cal_sharpe_ratio(return_series, rf):
    annual_return = cal_annual_return(return_series)
    vol = cal_annual_volatility(return_series)
    sharpe_ratio = (annual_return - rf) / vol

    return sharpe_ratio


"""
持仓的个股归因模块
"""
import pandas as pd
import hbshare as hbs
from hbshare.fe.common.util.config import style_name, industry_name
import datetime
import os
import numpy as np
from sqlalchemy import create_engine
from hbshare.quant.Kevin.rm_associated.config import engine_params
from hbshare.fe.common.util.config import factor_map_dict
from hbshare.quant.Kevin.quant_room.MyUtil.data_loader import get_daily_nav_from_work


class PrivateStockAttr:
    def __init__(self, fund_name, trade_date, end_date, benchmark_id):
        self.fund_name = fund_name
        self.trade_date = trade_date
        self.end_date = end_date
        self.benchmark_id = benchmark_id
        self._load_data()

    def _load_shift_date(self):
        trade_dt = datetime.datetime.strptime(self.trade_date, '%Y%m%d')
        pre_date = (trade_dt - datetime.timedelta(days=100)).strftime('%Y%m%d')

        sql_script = "SELECT JYRQ, SFJJ, SFZM, SFYM FROM funddb.JYRL WHERE JYRQ >= {} and JYRQ <= {}".format(
            pre_date, self.trade_date)
        res = hbs.db_data_query('readonly', sql_script, page_size=5000)
        df = pd.DataFrame(res['data']).rename(
            columns={"JYRQ": 'calendarDate', "SFJJ": 'isOpen',
                     "SFZM": "isWeekEnd", "SFYM": "isMonthEnd"}).sort_values(by='calendarDate')
        df['isOpen'] = df['isOpen'].astype(int).replace({0: 1, 1: 0})
        df['isWeekEnd'] = df['isWeekEnd'].fillna(0).astype(int)
        df['isMonthEnd'] = df['isMonthEnd'].fillna(0).astype(int)

        trading_day_list = df[df['isMonthEnd'] == 1]['calendarDate'].tolist()

        return trading_day_list[-1]

    def _load_portfolio_weight_series(self):
        sql_script = "SELECT * FROM private_fund_holding where fund_name = '{}' and trade_date = {}".format(
            self.fund_name, self.trade_date)
        engine = create_engine(engine_params)
        holding_df = pd.read_sql(sql_script, engine)
        holding_df['trade_date'] = holding_df['trade_date'].apply(lambda x: datetime.datetime.strftime(x, '%Y%m%d'))

        return holding_df.set_index('ticker')['weight'] / 100.

    def _load_benchmark_weight_series(self, date):
        sql_script = "SELECT * FROM hsjy_gg.SecuMain where SecuCategory = 4 and SecuCode = '{}'".format(
            self.benchmark_id)
        res = hbs.db_data_query('readonly', sql_script)
        index_info = pd.DataFrame(res['data'])
        inner_code = index_info.set_index('SECUCODE').loc[self.benchmark_id, 'INNERCODE']

        sql_script = "SELECT (select a.SecuCode from hsjy_gg.SecuMain a where a.InnerCode = b.InnerCode and " \
                     "rownum = 1) SecuCode, b.EndDate, b.Weight FROM hsjy_gg.LC_IndexComponentsWeight b WHERE " \
                     "b.IndexCode = '{}' and b.EndDate = to_date('{}', 'yyyymmdd')".format(inner_code, date)
        data = pd.DataFrame(hbs.db_data_query('readonly', sql_script)['data'])
        weight_df = data.rename(
            columns={"SECUCODE": "consTickerSymbol", "ENDDATE": "effDate", "WEIGHT": "weight"})

        return weight_df.set_index('consTickerSymbol')['weight'] / 100.

    @staticmethod
    def _load_stock_conception(date):
        sql_script = "SELECT * FROM stock_conception where trade_date = {}".format(date)
        engine = create_engine(engine_params)
        concept_df = pd.read_sql(sql_script, engine)
        concept_df['trade_date'] = concept_df['trade_date'].apply(lambda x: datetime.datetime.strftime(x, '%Y%m%d'))
        concept_df = concept_df[['trade_date', 'ticker', 'concept']]

        return concept_df

    @staticmethod
    def get_mkt_info(start_date, end_date):
        path = r'D:\MarketInfoSaver'
        listdir = os.listdir(path)
        listdir = [x for x in listdir if start_date < x.split('.')[0].split('_')[-1] <= end_date]
        data = []
        for filename in listdir:
            date = filename.split('.')[0].split('_')[-1]
            date_t_rate = pd.read_csv(os.path.join(path, filename), dtype={"tradeDate": str})
            date_t_rate['ticker'] = date_t_rate['ticker'].apply(lambda x: str(x).zfill(6))
            date_t_rate.loc[date_t_rate['turnoverValue'] < 1e-8, 'dailyReturnReinv'] = np.NAN
            date_t_rate['tradeDate'] = date
            data.append(date_t_rate)

        data = pd.concat(data)
        # filter
        data['dailyReturnReinv'] /= 100.
        # data = data[(data['dailyReturnReinv'] >= -0.2) & (data['dailyReturnReinv'] <= 0.2)]
        data = pd.pivot_table(data, index='tradeDate', columns='ticker', values='dailyReturnReinv').sort_index()

        return data

    @staticmethod
    def _load_sec_name(date):
        sql_script = "SELECT SYMBOL, SNAME FROM finchina.CHDQUOTE WHERE" \
                     " TDATE = {}".format(date)
        res = hbs.db_data_query('readonly', sql_script, page_size=6000)
        df = pd.DataFrame(res['data'])
        df = df[df['SYMBOL'].str[0].isin(['0', '3', '6'])]
        df.rename(columns={"SYMBOL": "ticker"}, inplace=True)

        return df.set_index('ticker')[['SNAME']]

    def _load_data(self):
        shift_date = self._load_shift_date()
        portfolio_weight_series = self._load_portfolio_weight_series()
        benchmark_weight_series = self._load_benchmark_weight_series(shift_date)
        concept_df = self._load_stock_conception(shift_date)
        mkt_info = self.get_mkt_info(self.trade_date, self.end_date)
        sec_name_df = self._load_sec_name(shift_date)

        self.data_param = {
            "portfolio_weight_series": portfolio_weight_series,
            "benchmark_weight_series": benchmark_weight_series,
            "concept_df": concept_df,
            "mkt_info": mkt_info,
            "sec_name_df": sec_name_df
        }

    def get_construct_result(self):
        portfolio_weight_series = self.data_param.get('portfolio_weight_series')
        benchmark_weight_series = self.data_param.get('benchmark_weight_series')
        benchmark_weight_series /= benchmark_weight_series.sum()
        weight_df = portfolio_weight_series.to_frame('port').merge(
            benchmark_weight_series.to_frame('bm'), left_index=True, right_index=True, how='outer').fillna(0.)
        # 没有IC的补正，用对应基准补齐
        weight_df['port'] += (1 - weight_df['port'].sum()) * weight_df['bm']
        period_data = self.data_param.get('mkt_info')
        # intersection
        idx = period_data.columns.intersection(weight_df.index)
        weight_df = weight_df.reindex(idx)
        period_data = period_data[idx].fillna(0.)
        period_data.loc[self.trade_date] = 0.
        period_data = (1 + period_data.sort_index()).cumprod()
        # 模拟净值
        simulate_nav = period_data.dot(weight_df)
        simulate_nav['trade_date'] = simulate_nav.index
        simulate_nav['trade_date'] = simulate_nav['trade_date'].apply(
            lambda x: datetime.datetime.strptime(x, "%Y%m%d").strftime("%Y-%m-%d"))
        simulate_nav.set_index('trade_date', inplace=True)
        # actual_nav = get_daily_nav_from_work(self.trade_date, self.end_date, {self.fund_name: "P50192"})
        # nav_df = simulate_nav.merge(actual_nav, left_index=True, right_index=True)
        # nav_df.plot.line()
        # 个股归因
        weight_df.eval("diff = port - bm", inplace=True)
        calc_df = weight_df.merge((period_data.loc[self.end_date] - 1).to_frame('period_ret'), left_index=True,
                                  right_index=True)
        calc_df.index.name = "ticker"
        calc_df.eval("excess_ret = diff * period_ret", inplace=True)
        sec_df = self.data_param.get("sec_name_df")
        calc_df = sec_df.merge(calc_df, left_index=True, right_index=True).reset_index()
        calc_df.loc[calc_df['bm'] == 0., 'cate'] = "基准外持仓"
        calc_df.loc[(calc_df['port'] >= calc_df['bm']) & (calc_df['bm'] > 0.), 'cate'] = "超配基准内部分"
        calc_df.loc[calc_df['port'] < calc_df['bm'], 'cate'] = "低配基准内部分"
        calc_df = calc_df.sort_values(by='excess_ret', ascending=False)
        cate_attr = calc_df.groupby('cate')[['port', 'excess_ret']].sum()
        # 概念归因
        concept_df = self.data_param.get('concept_df')[['ticker', 'concept']]
        concept_df = concept_df[['ticker']].join(pd.get_dummies(concept_df['concept']))
        concept_df = concept_df.groupby('ticker').sum()
        idx = set(calc_df['ticker']).intersection(concept_df.index)
        concept_df = concept_df.reindex(idx)
        concept_attr = concept_df.T.dot(calc_df.set_index('ticker').reindex(idx)[['diff', 'excess_ret']])
        concept_attr.rename(columns={"diff": "超低配", "excess_ret": "超额贡献"}, inplace=True)

        calc_df.rename(columns={"ticker": "股票代码", "SNAME": "股票名称", "port": "组合权重", "bm": "基准权重",
                                "period_ret": "区间收益", "excess_ret": "个股超额"}, inplace=True)

        excel_writer = pd.ExcelWriter('D:\\量化个股归因结果\\{}_{}-{}归因.xlsx'.format(
            self.fund_name, self.trade_date, self.end_date), engine='xlsxwriter')
        simulate_nav.rename(columns={"port": "组合", "bm": "基准"}).to_excel(excel_writer, sheet_name="模拟净值")
        cate_attr.applymap(lambda x: format(x, ".2%")).reset_index().rename(
            columns={"cate": "类别", "port": "权重", "excess_ret": "超额贡献"}).to_excel(
            excel_writer, sheet_name="超低配贡献", index=False)
        concept_attr.applymap(lambda x: format(x, ".2%")).to_excel(excel_writer, sheet_name="概念归因")

        del calc_df['cate']
        include_list = calc_df.select_dtypes(include=[np.number]).columns
        calc_df[include_list] = calc_df[include_list].applymap(lambda x: format(x, ".2%"))
        calc_df.to_excel(excel_writer, sheet_name="个股归因明细", index=False)

        excel_writer.close()


if __name__ == '__main__':
    PrivateStockAttr('千衍六和3号', '20230331', '20230421', '000905').get_construct_result()



"""
H1&N1的季度归因模块
"""
import os
import pandas as pd
import hbshare as hbs
from datetime import datetime


def calc_cate_attr_h1(data_path, s_date, e_date, date):
    # 逐月持仓
    holding_df = pd.read_excel(os.path.join(data_path, "H1-底表-{}.xlsx".format(date)), sheet_name="逐月持仓")
    holding_df.rename(columns={"Unnamed: 0": "type_1", "Unnamed: 1": "type_2",
                               "Unnamed: 2": "type_3", "Unnamed: 3": "name"}, inplace=True)
    holding_df.dropna(subset=['type_1', 'type_2'], inplace=True)
    holding_df.columns = [str(x) for x in holding_df.columns]
    include_list = ['type_1', 'type_2', 'type_3', 'name', s_date, e_date]
    holding_df = holding_df[include_list]
    # 交易信息
    trading_df = pd.read_excel(os.path.join(data_path, "H1-底表-{}.xlsx".format(date)), sheet_name="历史交易")
    trading_df = trading_df[trading_df.columns[:8]]
    trading_df = trading_df[['交易日期', '资产名称', '交易方向', '交易金额（万元）']]
    trading_df['交易日期'] = trading_df['交易日期'].apply(lambda x: datetime.strftime(x, "%Y%m%d"))
    trading_df = trading_df[(trading_df['交易日期'] > s_date) & (trading_df['交易日期'] < e_date)]
    # set1 = set(holding_df['name'])
    # set2 = set(trading_df['资产名称'])
    # set_minus = set2 - set1
    map_dict = {
        "乾象股票对冲3号私募证券投资基金": "乾象股票对冲3号私募证券投资基金B类",
        "卓识利民二号私募证券投资基金": "卓识利民二号私募证券投资基金A类",
        "景顺长城环保优势（001975）": "景顺长城环保优势股票",
        "正松云睿全球私募证券投资基金": "正松云睿全球私募证券投资基金A类",
        "浙商智选价值C（010382.OF）": "浙商智选价值混合C",
        "诚奇睿盈对冲私募证券投资基金": "诚奇睿盈对冲私募证券投资基金A类",
    }
    trading_df['资产名称'].replace(map_dict, inplace=True)
    trading_df.rename(columns={"资产名称": "name"}, inplace=True)
    pivot_tr = pd.pivot_table(trading_df, index='name', columns="交易方向", values='交易金额（万元）').fillna(0.)
    pivot_tr *= 10000.

    df = holding_df.merge(pivot_tr, left_on='name', right_index=True, how='left').fillna(0.)
    # special treat
    df.loc[df['name'] == '英仕曼宏量1号私募基金C类（单外包）', e_date] = 0.
    df.rename(columns={s_date: "起始日期", e_date: "结束日期"}, inplace=True)
    df.eval("pnl = 结束日期 - 起始日期 + 赎回 - 申购", inplace=True)
    df.eval("市值 = 结束日期 + 赎回", inplace=True)
    df.loc[df['type_2'] == "股票", 'type_2'] = df['type_3']
    attr_df = df.groupby(['type_1', 'type_2'])['pnl', '市值'].sum().reset_index()
    attr_df['收益率'] = attr_df['pnl'] / attr_df['市值']
    attr_df.rename(columns={"type_1": "超一级", "type_2": "一级", "pnl": "盈亏"}, inplace=True)

    s_date = "20221230"
    sql_script = "SELECT * FROM st_hedge.t_st_sm_zhmzs WHERE zsdm in ('HB0011', 'HB1001', 'HB1002', 'HB0015', 'HB0018') and " \
                 "jyrq >= {} and jyrq <= {}".format(s_date, e_date)
    res = hbs.db_data_query('highuser', sql_script)
    data = pd.DataFrame(res['data'])
    hb_index = data.pivot_table(index='jyrq', columns='zsdm', values='spjg').sort_index()
    bm_ret = hb_index.loc[e_date] / hb_index.loc[s_date] - 1

    return attr_df, bm_ret


def calc_cate_attr_n1(data_path, s_date, e_date, date):
    # 逐月持仓
    holding_df = pd.read_excel(os.path.join(data_path, "N1-底表-{}.xlsx".format(date)), sheet_name="逐月持仓")
    holding_df.rename(columns={"Unnamed: 0": "type_1", "Unnamed: 1": "type_2",
                               "Unnamed: 2": "type_3", "Unnamed: 3": "name"}, inplace=True)
    holding_df.dropna(subset=['type_1', 'type_2'], inplace=True)
    holding_df.columns = [str(x) for x in holding_df.columns]
    include_list = ['type_1', 'type_2', 'type_3', 'name', s_date, e_date]
    holding_df = holding_df[include_list]
    # 交易信息
    trading_df = pd.read_excel(os.path.join(data_path, "N1-底表-{}.xlsx".format(date)), sheet_name="历史交易")
    trading_df = trading_df[trading_df.columns[:9]]
    trading_df = trading_df[['交易日期', '投资基金', '投资类型', '交易金额(万元)']]
    trading_df['交易日期'] = trading_df['交易日期'].apply(lambda x: datetime.strftime(x, "%Y%m%d"))
    trading_df = trading_df[(trading_df['交易日期'] > s_date) & (trading_df['交易日期'] < e_date)]
    # set1 = set(holding_df['name'])
    # set2 = set(trading_df['投资基金'])
    # set_minus = set2 - set1
    map_dict = {
        "国富中小盘": "国富中小盘股票",
    }
    trading_df['投资基金'].replace(map_dict, inplace=True)
    trading_df.rename(columns={"投资基金": "name"}, inplace=True)
    trading_df = trading_df.groupby(['name', '投资类型'])['交易金额(万元)'].sum().reset_index()
    pivot_tr = pd.pivot_table(trading_df, index='name', columns="投资类型", values='交易金额(万元)').fillna(0.)
    pivot_tr *= 10000.

    df = holding_df.merge(pivot_tr, left_on='name', right_index=True, how='left').fillna(0.)
    # special treat
    # df.loc[df['name'] == '勤辰森裕2号私募证券投资基金', e_date] = 4361143.26
    # df.loc[df['name'] == '易方达供给改革', e_date] = 4361143.26
    df.rename(columns={s_date: "起始日期", e_date: "结束日期"}, inplace=True)
    df.eval("pnl = 结束日期 - 起始日期 + 赎回 - 申购", inplace=True)
    df.eval("市值 = 结束日期 + 赎回", inplace=True)
    df.loc[df['type_2'] == "股票", 'type_2'] = df['type_3']
    attr_df = df.groupby(['type_1', 'type_2'])['pnl', '市值'].sum().reset_index()
    attr_df['收益率'] = attr_df['pnl'] / attr_df['市值']
    attr_df.rename(columns={"type_1": "超一级", "type_2": "一级", "pnl": "盈亏"}, inplace=True)

    s_date = "20221230"
    sql_script = "SELECT * FROM st_hedge.t_st_sm_zhmzs WHERE zsdm in ('HB0011', 'HB1001', 'HB1002', 'HB0015', 'HB0018') and " \
                 "jyrq >= {} and jyrq <= {}".format(s_date, e_date)
    res = hbs.db_data_query('highuser', sql_script)
    data = pd.DataFrame(res['data'])
    hb_index = data.pivot_table(index='jyrq', columns='zsdm', values='spjg').sort_index()
    bm_ret = hb_index.loc[e_date] / hb_index.loc[s_date] - 1

    return attr_df, bm_ret


if __name__ == '__main__':
    path = "D:\\研究基地\\Y-H1&N1\\底表"
    start_date = "20221231"
    end_date = "20230331"
    trade_date = "20230414"
    # calc_cate_attr_h1(path, start_date, end_date, trade_date)
    calc_cate_attr_n1(path, start_date, end_date, trade_date)



"""
Alpha月报
"""
import os
import numpy as np
import pandas as pd
import hbshare as hbs
import datetime
import math
from scipy import stats
from sqlalchemy import create_engine
from hbshare.quant.Kevin.quant_room.MyUtil.util_func import run_time
from hbshare.quant.Kevin.rm_associated.config import engine_params, style_names
from hbshare.quant.Kevin.quant_room.MyUtil.data_loader import get_fund_nav_from_sql, get_trading_day_list
from periodic_config import index_name_dict, industry_name_dict, style_name_dict, range_start_dict
from hbshare.quant.CChen.cons import sql_write_path_hb
from hbshare.quant.CChen.fut import wind_stk_index_basis
from WindPy import w
import warnings

warnings.filterwarnings("ignore")
w.start()


class MonthlyReporter:
    def __init__(self, trade_date):
        self.trade_date = trade_date
        self._date_preprocess()

    def _date_preprocess(self):
        trade_dt = datetime.datetime.strptime(self.trade_date, '%Y%m%d')
        range_start = (trade_dt - datetime.timedelta(days=500)).strftime('%Y%m%d')
        month_list = get_trading_day_list(range_start, self.trade_date, frequency="month")
        self.pre_date = month_list[month_list.index(self.trade_date) - 1]
        self.start_date = [x for x in month_list if x[4:6] == "12"][-1]
        self.trading_day_list_tm = get_trading_day_list(self.pre_date, self.trade_date)
        self.trading_day_list_ty = get_trading_day_list(self.start_date, self.trade_date)

    def _calc_market_index_ret(self):
        """
        指数收益
        """
        index_list = list(index_name_dict.keys())
        res = w.wsd(','.join(index_list), "close", self.start_date, self.trade_date)
        if res.ErrorCode != 0:
            data = pd.DataFrame()
            print("fetch market index data error: start_date = {}, end_date = {}".format(
                self.pre_date, self.trade_date))
        else:
            data = pd.DataFrame(res.Data, index=res.Codes, columns=res.Times).T
            data.index.name = 'trade_date'
            data.reset_index(inplace=True)
            data['trade_date'] = data['trade_date'].apply(lambda x: datetime.datetime.strftime(x, '%Y%m%d'))
            data = data.set_index('trade_date').rename(columns=index_name_dict).sort_index().dropna()
            data /= data.iloc[0]
            assert (data.shape[0] == len(self.trading_day_list_ty))

        index_period = data.loc[self.pre_date: self.trade_date]
        index_period /= index_period.iloc[0]
        index_period['trade_date'] = index_period.index
        index_period['trade_date'] = index_period['trade_date'].apply(
            lambda x: datetime.datetime.strptime(x, "%Y%m%d").strftime("%m-%d"))
        index_period.set_index('trade_date', inplace=True)

        M_ret = data.loc[self.trade_date] / data.loc[self.pre_date] - 1
        Ytd_ret = data.loc[self.trade_date] / data.loc[self.start_date] - 1
        ret_df = M_ret.to_frame("本月").merge(Ytd_ret.to_frame('今年以来'), left_index=True, right_index=True)

        return index_period, ret_df

    def _calc_industry_ret(self):
        """
        行业收益及成交情况
        """
        industry_list = list(industry_name_dict.keys())
        # 行业指数收益
        res = w.wsd(','.join(industry_list), "close", self.pre_date, self.trade_date)
        if res.ErrorCode != 0:
            data = pd.DataFrame()
            print("fetch industry return data error: start_date = {}, end_date = {}".format(
                self.pre_date, self.trade_date))
        else:
            data = pd.DataFrame(res.Data, index=res.Codes, columns=res.Times).T
            data.index.name = 'trade_date'
            data.reset_index(inplace=True)
            data['trade_date'] = data['trade_date'].apply(lambda x: datetime.datetime.strftime(x, '%Y%m%d'))
            data = data.set_index('trade_date').rename(columns=industry_name_dict).sort_index().dropna()
            data /= data.iloc[0]
            assert (data.shape[0] == len(self.trading_day_list_tm))
        industry_ret = data.loc[self.trade_date].sort_values() - 1
        # 行业成交额
        res = w.wsd(','.join(industry_list), "amt", self.pre_date, self.trade_date)
        if res.ErrorCode != 0:
            data = pd.DataFrame()
            print("fetch industry amt data error: start_date = {}, end_date = {}".format(
                self.pre_date, self.trade_date))
        else:
            data = pd.DataFrame(res.Data, index=res.Codes, columns=res.Times).T
            data.index.name = 'trade_date'
            data.reset_index(inplace=True)
            data['trade_date'] = data['trade_date'].apply(lambda x: datetime.datetime.strftime(x, '%Y%m%d'))
            data = data.set_index('trade_date').rename(columns=industry_name_dict).sort_index().dropna()
            assert (data.shape[0] == len(self.trading_day_list_tm))
        industry_amt = (data.sum() / data.sum().sum()).sort_values()

        industry_data = industry_ret.to_frame('月度涨跌幅').merge(
            industry_amt.to_frame('月度成交占比'), left_index=True, right_index=True)

        return industry_data.sort_values(by='月度涨跌幅')

    def _calc_style_factor_ret(self):
        """
        风格因子收益
        """
        sql_script = "SELECT * FROM factor_return where " \
                     "TRADE_DATE > {} and TRADE_DATE <= {}".format(self.pre_date, self.trade_date)
        engine = create_engine(engine_params)
        factor_return = pd.read_sql(sql_script, engine)
        factor_return['trade_date'] = factor_return['trade_date'].apply(lambda x: datetime.datetime.strftime(x, '%Y%m%d'))
        factor_return = pd.pivot_table(
            factor_return, index='trade_date', columns='factor_name', values='factor_ret').sort_index()[style_names]
        assert (factor_return.shape[0] == len(self.trading_day_list_tm) - 1)
        factor_return.loc[self.pre_date] = 0.
        factor_return.sort_index(inplace=True)

        factor_return['trade_date'] = factor_return.index
        factor_return['trade_date'] = factor_return['trade_date'].apply(
            lambda x: datetime.datetime.strptime(x, "%Y%m%d").strftime("%m-%d"))
        factor_return.set_index('trade_date', inplace=True)
        factor_return.rename(columns=style_name_dict, inplace=True)

        return factor_return.cumsum(), factor_return.sum().to_frame('本月')

    def _calc_trading_liquidity(self):
        """
        指数成交活跃度
        """
        start_date = range_start_dict['指数成交活跃度']
        sql_script = "SELECT * FROM mac_stock_trading WHERE TRADE_DATE > {} and TRADE_DATE <= {}".format(
            start_date, self.trade_date)
        engine = create_engine(engine_params)
        data = pd.read_sql(sql_script, engine)
        data['trade_date'] = data['trade_date'].apply(lambda x: datetime.datetime.strftime(x, '%Y%m%d'))
        amt_data = data.set_index('trade_date')[['amt_300', 'amt_500', 'amt_1000', 'amt_other']].sort_index()
        amt_ratio = amt_data.div(amt_data.sum(axis=1), axis=0)
        amt_ratio['trade_date'] = amt_ratio.index
        amt_ratio['trade_date'] = amt_ratio['trade_date'].apply(
            lambda x: datetime.datetime.strptime(x, "%Y%m%d").strftime("%Y-%m-%d"))
        amt_ratio.set_index('trade_date', inplace=True)
        amt_ratio.rename(columns={"amt_300": "沪深300", "amt_500": "中证500",
                                  "amt_1000": "中证1000", "amt_other": "剩余小票"}, inplace=True)

        return amt_ratio

    def _calc_valuation(self):
        """
        指数估值数据
        """
        trade_dt = datetime.datetime.strptime(self.trade_date, '%Y%m%d')
        range_start = (trade_dt - datetime.timedelta(days=365*12)).strftime('%Y%m%d')
        sql_script = "SELECT * FROM mac_stock_pe_ttm where TRADE_DATE >= {} and TRADE_DATE <= {}".format(
            range_start, self.trade_date)
        engine = create_engine(engine_params)
        pe_ttm = pd.read_sql(sql_script, engine)
        pe_ttm['trade_date'] = pe_ttm['trade_date'].apply(lambda x: datetime.datetime.strftime(x, '%Y%m%d'))
        pe_ttm = pe_ttm.set_index('trade_date').rename(
            columns={"SZ50": "上证50", "HS300": "沪深300", "ZZ500": "中证500", "ZZ1000": "中证1000"})
        pe_ttm = pe_ttm[['上证50', '沪深300', '中证500', '中证1000']].sort_index()
        # 过往十年的历史分位
        pe_quantile = pe_ttm.rolling(
            252 * 10, min_periods=252 * 5).apply(lambda x: x.dropna().rank(pct=True).iloc[-1]).dropna()
        month_list = get_trading_day_list(range_start_dict['估值分位'], self.trade_date, frequency="month")
        pe_quantile = pe_quantile.reindex(month_list)
        pe_quantile['trade_date'] = pe_quantile.index
        pe_quantile['trade_date'] = pe_quantile['trade_date'].apply(
            lambda x: datetime.datetime.strptime(x, "%Y%m%d").strftime("%Y-%m"))
        pe_quantile.set_index('trade_date', inplace=True)

        return pe_quantile

    @staticmethod
    def _calc_stock_bond_premium(benchmark_name):
        # 宽基指数股息率
        sql_script = "SELECT trade_date, {} FROM mac_stock_dividend".format(benchmark_name)
        engine = create_engine(engine_params)
        dividend = pd.read_sql(sql_script, engine).dropna()
        # 国债收益率数据
        sql_script = "SELECT trade_date, ytm_10y FROM mac_treasury_yield"
        treasury_data = pd.read_sql(sql_script, engine)
        data = pd.merge(dividend, treasury_data, on='trade_date').dropna()
        data['trade_date'] = data['trade_date'].apply(lambda x: datetime.datetime.strftime(x, '%Y%m%d'))
        # 指数行情数据
        map_dict = {"SZZS": "000001", "SZ50": "000016", "HS300": "000300", "ZZ500": "000905", "ZZ1000": "000852"}
        sql_script = "SELECT JYRQ, ZQDM, SPJG FROM funddb.ZSJY where ZQDM = '{}' and JYRQ >= '20080101'".format(
            map_dict[benchmark_name])
        res = hbs.db_data_query('readonly', sql_script, page_size=5000)
        index_df = pd.DataFrame(res['data']).rename(columns={"JYRQ": "trade_date", "SPJG": "benchmark"})
        data = pd.merge(data, index_df[['trade_date', 'benchmark']], on='trade_date').sort_values(by='trade_date')

        data['premium'] = (data['ytm_10y'] - data[benchmark_name]) / 100.
        window = 750
        data['roll_mean'] = data['premium'].rolling(window).mean()
        data['roll_std'] = data['premium'].rolling(window).std()
        data = data.dropna()
        data['trade_date'] = data['trade_date'].apply(
            lambda x: datetime.datetime.strptime(x, "%Y%m%d").strftime("%Y-%m-%d"))
        data.set_index('trade_date', inplace=True)
        data['+1 std'] = data['roll_mean'] + data['roll_std']
        data['+2 std'] = data['roll_mean'] + 2 * data['roll_std']
        data['-1 std'] = data['roll_mean'] - data['roll_std']
        data['-2 std'] = data['roll_mean'] - 2 * data['roll_std']
        d = {"HS300": "沪深300", "ZZ500": "中证500"}
        data.rename(columns={
            "benchmark": d[benchmark_name], "premium": "股债收益差", "roll_mean": "股债收益差均值"}, inplace=True)

        data = data[[d[benchmark_name], '股债收益差', '股债收益差均值', '+1 std', '+2 std', '-1 std', '-2 std']]

        return data

    def _calc_index_basis(self):
        range_start = range_start_dict['股指基差']
        start_date = datetime.datetime.strptime(range_start, '%Y%m%d').date()
        end_date = datetime.datetime.strptime(self.trade_date, '%Y%m%d').date()
        sql_path = sql_write_path_hb['daily']
        include_cols = ['t_date', 'symbol', 'basis']
        if_df = wind_stk_index_basis(code='IF', start_date=start_date, end_date=end_date,
                                     sql_path=sql_path, table="futures_wind")[2][include_cols]
        ic_df = wind_stk_index_basis(code='IC', start_date=start_date, end_date=end_date,
                                     sql_path=sql_path, table="futures_wind")[2][include_cols]
        im_df = wind_stk_index_basis(code='IM', start_date=start_date, end_date=end_date,
                                     sql_path=sql_path, table="futures_wind")[2][include_cols]
        basis_df = pd.concat([if_df, ic_df, im_df])
        basis_df['t_date'] = basis_df['t_date'].apply(lambda x: datetime.datetime.strftime(x, '%Y%m%d'))
        basis_df = basis_df[basis_df['t_date'] > range_start]
        basis_df = pd.pivot_table(basis_df, index='t_date', columns='symbol', values='basis').sort_index()
        basis_df['trade_date'] = basis_df.index
        basis_df['trade_date'] = basis_df['trade_date'].apply(
            lambda x: datetime.datetime.strptime(x, "%Y%m%d").strftime("%Y-%m-%d"))
        basis_df.set_index('trade_date', inplace=True)
        basis_df.rename(columns={"IF02.CFE": "IF当季", "IC02.CFE": "IC当季", "IM02.CFE": "IM当季"}, inplace=True)
        basis_df /= 100.

        return basis_df

    def _calc_market_trading(self):
        range_start = range_start_dict['市场成交额']
        sql_script = "SELECT * FROM mac_stock_trading WHERE TRADE_DATE > {} and TRADE_DATE <= {}".format(
            range_start, self.trade_date)
        engine = create_engine(engine_params)
        data = pd.read_sql(sql_script, engine)
        data['trade_date'] = data['trade_date'].apply(lambda x: datetime.datetime.strftime(x, '%Y%m%d'))
        amt_data = data[['trade_date', 'amt_sh', 'amt_sz']].sort_values(by='trade_date')
        amt_data['trade_date'] = amt_data['trade_date'].apply(
            lambda x: datetime.datetime.strptime(x, "%Y%m%d").strftime("%Y-%m-%d"))
        amt_data.set_index('trade_date', inplace=True)
        amt_data.rename(columns={"amt_sh": "上证", "amt_sz": "深证"}, inplace=True)
        amt_data.eval("全市场 = 上证 + 深证", inplace=True)

        return amt_data

    def _calc_trading_cr(self):
        range_start = range_start_dict['成交集中度']
        sql_script = "SELECT * FROM mac_stock_trading_cr WHERE TRADE_DATE > {} and TRADE_DATE <= {}".format(
            range_start, self.trade_date)
        engine = create_engine(engine_params)
        data = pd.read_sql(sql_script, engine)
        data['trade_date'] = data['trade_date'].apply(lambda x: datetime.datetime.strftime(x, '%Y-%m-%d'))
        data = data.set_index('trade_date')[['cr5']].sort_index()
        data.rename(columns={"cr5": "日成交额前5%占比"}, inplace=True)

        return data

    def _calc_stock_diff(self):
        start_date = range_start_dict['截面波动率']
        trade_dt = datetime.datetime.strptime(start_date, '%Y%m%d')
        range_start = (trade_dt - datetime.timedelta(days=100)).strftime('%Y%m%d')
        path = r'D:\MarketInfoSaver'
        listdir = os.listdir(path)
        listdir = [x for x in listdir if range_start <= x.split('.')[0].split('_')[-1] <= self.trade_date]
        date_list = []
        vol = []
        for filename in listdir:
            date = filename.split('.')[0].split('_')[-1]
            date_list.append(date)
            date_t_rate = pd.read_csv(os.path.join(path, filename), dtype={"tradeDate": str})
            date_t_rate['ticker'] = date_t_rate['ticker'].apply(lambda x: str(x).zfill(6))
            date_t_rate.loc[date_t_rate['turnoverValue'] < 1e-8, 'dailyReturnReinv'] = np.NAN
            date_t_rate = date_t_rate[date_t_rate['dailyReturnReinv'].abs() <= 20]
            vol.append(date_t_rate['dailyReturnReinv'].dropna().std() / 100.)

        vol_df = pd.Series(index=date_list, data=vol).to_frame('stock_diff').sort_index()
        vol_df['mean_10'] = vol_df['stock_diff'].rolling(10).mean()
        vol_df = vol_df.loc[start_date:].dropna()
        vol_df.rename(columns={"stock_diff": "截面波动率", "mean_10": "10日平均"}, inplace=True)

        vol_df['trade_date'] = vol_df.index
        vol_df['trade_date'] = vol_df['trade_date'].apply(
            lambda x: datetime.datetime.strptime(x, "%Y%m%d").strftime("%Y-%m-%d"))
        vol_df.set_index('trade_date', inplace=True)

        return vol_df

    @staticmethod
    def get_mkt_info(start_date, end_date):
        path = r'D:\MarketInfoSaver'
        listdir = os.listdir(path)
        listdir = [x for x in listdir if start_date < x.split('.')[0].split('_')[-1] <= end_date]
        data = []
        for filename in listdir:
            date = filename.split('.')[0].split('_')[-1]
            date_t_rate = pd.read_csv(os.path.join(path, filename), dtype={"tradeDate": str})
            date_t_rate['ticker'] = date_t_rate['ticker'].apply(lambda x: str(x).zfill(6))
            date_t_rate.loc[date_t_rate['turnoverValue'] < 1e-8, 'dailyReturnReinv'] = np.NAN
            date_t_rate['tradeDate'] = date
            data.append(date_t_rate)

        data = pd.concat(data)
        # filter
        data['dailyReturnReinv'] /= 100.
        data = data[(data['dailyReturnReinv'] >= -0.2) & (data['dailyReturnReinv'] <= 0.2)]
        data = pd.pivot_table(data, index='tradeDate', columns='ticker', values='dailyReturnReinv').sort_index()
        data = data.dropna(how='any', axis=1)

        return (1 + data).prod() - 1

    def _calc_win_ratio(self):
        range_start = range_start_dict['成交集中度']
        week_list = get_trading_day_list(range_start, self.trade_date, frequency="week")
        sql_script = "SELECT JYRQ as TRADEDATE, ZQDM, SPJG as TCLOSE from funddb.ZSJY WHERE " \
                     "ZQDM in ('000300','000905','000852') " \
                     "and JYRQ >= {} and JYRQ <= {}".format(range_start, self.trade_date)
        data = pd.DataFrame(hbs.db_data_query('readonly', sql_script, page_size=5000)['data'])
        index_return = data.pivot_table(
            index='TRADEDATE', columns='ZQDM', values='TCLOSE').reindex(week_list).pct_change().dropna()

        win_ratio_df = pd.DataFrame(index=index_return.index, columns=index_return.columns)
        for i in range(len(week_list) - 1):
            pre_week = week_list[i]
            current_week = week_list[i + 1]
            period_return = self.get_mkt_info(pre_week, current_week)
            index_ret = index_return.loc[current_week]
            win_ratio = 100 - stats.percentileofscore(np.array(period_return), index_ret)
            win_ratio_df.loc[current_week] = win_ratio
        win_ratio_df.rename(columns={"000300": "沪深300", "000905": "中证500", "000852": "中证1000"}, inplace=True)

        win_ratio_df['trade_date'] = win_ratio_df.index
        win_ratio_df['trade_date'] = win_ratio_df['trade_date'].apply(
            lambda x: datetime.datetime.strptime(x, "%Y%m%d").strftime("%Y-%m-%d"))
        win_ratio_df.set_index('trade_date', inplace=True)

        return win_ratio_df / 100.

    @run_time
    def run(self):
        excel_writer = pd.ExcelWriter(
            'D:\\Alpha_Monthly_Report\\alpha_data\\月报数据.xlsx', engine='xlsxwriter')
        # 市场概况
        index_period, ret_df = self._calc_market_index_ret()
        index_period.to_excel(excel_writer, sheet_name='主要指数走势')
        ret_df.to_excel(excel_writer, sheet_name="指数区间涨跌幅")
        industry_data = self._calc_industry_ret()
        industry_data.to_excel(excel_writer, sheet_name="行业涨跌幅&成交占比")
        style_trend, style_ret = self._calc_style_factor_ret()
        style_trend.to_excel(excel_writer, sheet_name="风格因子收益率走势")
        style_ret.to_excel(excel_writer, sheet_name="风格因子涨跌幅")

        # beta环境
        trading_liquidity = self._calc_trading_liquidity()
        trading_liquidity.to_excel(excel_writer, sheet_name="市场流动性分布")
        pe_quantile = self._calc_valuation()
        pe_quantile.to_excel(excel_writer, sheet_name="PE估值分位")
        premium_300 = self._calc_stock_bond_premium("HS300")
        premium_500 = self._calc_stock_bond_premium("ZZ500")
        premium_300.to_excel(excel_writer, sheet_name="沪深300-股债收益差")
        premium_500.to_excel(excel_writer, sheet_name="中证500-股债收益差")
        index_basis = self._calc_index_basis()
        index_basis.to_excel(excel_writer, sheet_name="股指期货基差走势")

        # alpha微观结构
        trading_df = self._calc_market_trading()
        trading_df.to_excel(excel_writer, sheet_name="市场成交额")
        trading_cr = self._calc_trading_cr()
        trading_cr.to_excel(excel_writer, sheet_name="成交集中度")
        stock_diff = self._calc_stock_diff()
        stock_diff.to_excel(excel_writer, sheet_name="截面波动率")
        index_win_ratio = self._calc_win_ratio()
        index_win_ratio.to_excel(excel_writer, sheet_name="战胜指数比例")

        excel_writer.close()


if __name__ == '__main__':
    MonthlyReporter(trade_date="20230331").run()


"""
概念类指数权重入库模块：Wind小市值指数
"""
import pandas as pd
import datetime
from hbshare.quant.Kevin.asset_allocation.macro_index.util import create_table, delete_duplicate_records, WriteToDB
from WindPy import w

w.start()


class BenchmarkHoldingExtractor:
    def __init__(self, trade_date, table_name, benchmark_id, is_increment=1):
        self.trade_date = trade_date
        self.table_name = table_name
        self.benchmark_id = benchmark_id
        self.is_increment = is_increment

    def _load_weight_from_wind(self):
        trade_date = datetime.datetime.strptime(self.trade_date, '%Y%m%d').strftime('%Y%m%d')
        res = w.wset("indexconstituent", "date={};windcode={}".format(trade_date, self.benchmark_id))
        data = pd.DataFrame(data=res.Data, index=res.Fields).T
        data.rename(columns={"date": "trade_date", "wind_code": "ticker", "i_weight": "weight"}, inplace=True)
        data['trade_date'] = data['trade_date'].apply(lambda x: datetime.datetime.strftime(x, '%Y%m%d'))
        data['ticker'] = data['ticker'].apply(lambda x: x.split('.')[0])
        del data['industry']
        data['benchmark_id'] = self.benchmark_id

        return data

    def writeToDB(self):
        if self.is_increment == 1:
            data = self._load_weight_from_wind()
            trading_day_list = data['trade_date'].unique().tolist()
            sql_script = "delete from {} where trade_date in ({}) and benchmark_id = '{}'".format(
                self.table_name, ','.join(trading_day_list), self.benchmark_id)
            # delete first
            delete_duplicate_records(sql_script)
            # add new records
            WriteToDB().write_to_db(data, self.table_name)
        else:
            sql_script = """
                create table {}(
                id int auto_increment primary key,
                trade_date date not null,
                ticker varchar(10),
                sec_name varchar(20),
                weight decimal(7, 6),
                benchmark_id varchar(20))
            """.format(self.table_name)
            create_table(self.table_name, sql_script)
            data = self._load_weight_from_wind()
            WriteToDB().write_to_db(data, self.table_name)


if __name__ == '__main__':
    name_outer = '衍复小市值指数增强一号'
    BenchmarkHoldingExtractor("20221230", "benchmark_holding", "8841425.WI", is_increment=1).writeToDB()